@@ -9,9 +9,12 @@ import numpy as np
 from dataclasses import dataclass
 from opt_einsum import contract
 
-import psi4
-psi4.core.be_quiet()
-psi4.core.clean()
+try:
+    import psi4
+    psi4.core.be_quiet()
+    psi4.core.clean()
+except:
+    pass    
 
 from .methods.wuyang import WuYang
 from .methods.zmp import ZMP
@@ -51,10 +54,10 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
         Reference calculation
         1 -> Restricted
         2 -> Unrestricted
-    nt : List
-        List of psi4.core.Matrix for target densities
+    Dt : List
+        List of psi4.core.Matrix for target density matrices (on AO).
     ct : List
-        List of psi4.core.Matrix for occupied orbitals
+        List of psi4.core.Matrix for input occupied orbitals. This might not be correct for post-HartreeFock methods.
     pbs_str: string
         name of Potential basis set
     pbs : psi4.core.BasisSet
@@ -116,15 +119,14 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
         self.ref       = 1 if psi4.core.get_global_option("REFERENCE") == "RHF" or \
                               psi4.core.get_global_option("REFERENCE") == "RKS" else 2
         self.jk        = wfn.jk() if hasattr(wfn, "jk") == True else self.generate_jk()
-        self.nt        = (np.array(wfn.Da_subset("AO")), np.array(wfn.Db_subset("AO")))
+        self.Dt        = (np.array(wfn.Da_subset("AO")), np.array(wfn.Db_subset("AO")))
         self.ct        = (np.array(wfn.Ca_subset("AO", "OCC")), np.array(wfn.Cb_subset("AO", "OCC")))
         self.pbs       = self.basis if pbs == "same" \
                                     else psi4.core.BasisSet.build( self.mol, key='BASIS', target=self.pbs_str)
         self.npbs      = self.pbs.nbf()
-        self.v_pbs        = np.zeros( (self.npbs) ) if self.ref == 1 \
+        self.v_pbs     = np.zeros( (self.npbs) ) if self.ref == 1 \
                                                  else np.zeros( 2 * self.npbs )
         self.generate_mints_matrices()
-
         self.grid = data_bucket
         self.cubic_grid = data_bucket
         
@@ -286,19 +288,38 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
             ----------
                 lambda_list: list
                     List of Lamda parameters used as a coefficient for Hartree 
-                    difference in SCF cycle. 
-                zmp_functional: str
-                    Specifies what functional to use to drive the SCF procedure.
-                    Options: {'hartree', 'log', 'exp', 'grad'}
-                zmp_mixing: float
-                    mixing \in [0,1]. How much of the new potential is added in 
-                    a given scf step. Values close to 1 may prevent SCF to convergence. 
+                    difference in SCF cycle.
+                zmp_mixing: float, optional
+                    mixing \in [0,1]. How much of the new potential is added in.
+                    For example, zmp_mixing = 0 means the traditional ZMP, i.e. all the potentials from previous
+                    smaller lambda are ignored.
+                    Zmp_mixing = 1 means that all the potentials of previous lambdas are accumulated, the larger lambda
+                    potential are meant to fix the wrong/inaccurate region of the potential of the sum of the previous
+                    potentials instead of providing an entire new potentials.
+                    default: 1
                 opt_max_iter: float
                     Maximum number of iterations for scf cycle
                 opt_tol: float
                     Convergence criteria set for Density Difference and DIIS error. 
                 return:
-                    The result will be stored in self.grid.vxc
+                    The result will be stored in self.proto_density_a and self.proto_density_b
+                    For zmp_mixing==1, restricted (ref==1):
+                        self.proto_density_a = \sum_i lambda_i * (Da_i - Dt[0]) - 1/N * (Dt[0])
+                        self.proto_density_b = \sum_i lambda_i * (Db_i - Dt[1]) - 1/N * (Dt[1]);
+                    unrestricted (ref==1):
+                        self.proto_density_a = \sum_i lambda_i * (Da_i - Dt[0]) - 1/N * (Dt[0] + Dt[1])
+                        self.proto_density_b = \sum_i lambda_i * (Db_i - Dt[1]) - 1/N * (Dt[0] + Dt[1]);
+                    For restricted (ref==1):
+                        vxc = \int dr' \frac{self.proto_density_a + self.proto_density_b}{|r-r'|}
+                            = 2 * \int dr' \frac{self.proto_density_a}{|r-r'|};
+                    for unrestricted (ref==2):
+                        vxc_up = \int dr' \frac{self.proto_density_a}{|r-r'|}
+                        vxc_down = \int dr' \frac{self.proto_density_b}{|r-r'|}.
+                    To get potential on grid, one needs to do
+                        vxc = self.on_grid_esp(Da=self.proto_density_a, Db=self.proto_density_b, grid=grid) for restricted;
+                        vxc_up = self.on_grid_esp(Da=self.proto_density_a, Db=np.zeros_like(self.proto_density_a),
+                                  grid=grid) for unrestricted;
+
 
         mRKS
         ----
@@ -459,7 +480,7 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
                 self.wfn.Da().diagonalize(C_NO, eigs_NO, psi4.core.DiagonalizeOrder.Descending)
                 occu = np.sqrt(eigs_NO.np)
                 New_Orb_a = occu * C_NO.np
-                assert np.allclose(New_Orb_a @ New_Orb_a.T, self.nt[0])
+                assert np.allclose(New_Orb_a @ New_Orb_a.T, self.Dt[0])
                 if self.ref == 1:
                     New_Orb_b = New_Orb_a
                 else:
@@ -489,15 +510,15 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
                 _, wfn_0 = psi4.energy( "svwn"+"/"+self.basis_str, molecule=self.mol , return_wfn = True)
 
             if self.ref == 1:
-                ntarget = psi4.core.Matrix.from_array( [ self.nt[0] + self.nt[1] ] )
+                ntarget = psi4.core.Matrix.from_array( [ self.Dt[0] + self.Dt[1] ] )
                 wfn_0.V_potential().set_D( [ntarget] )
                 vxc_a = psi4.core.Matrix( self.nbf, self.nbf )
                 wfn_0.V_potential().compute_V([vxc_a])
                 self.va += vxc_a.np
                 self.vb += vxc_a.np
             elif self.ref == 2:
-                na_target = psi4.core.Matrix.from_array( self.nt[0] )
-                nb_target = psi4.core.Matrix.from_array( self.nt[1] )
+                na_target = psi4.core.Matrix.from_array( self.Dt[0] )
+                nb_target = psi4.core.Matrix.from_array( self.Dt[1] )
                 wfn_0.V_potential().set_D( [na_target, nb_target] )
                 vxc_a = psi4.core.Matrix( self.nbf, self.nbf )
                 vxc_b = psi4.core.Matrix( self.nbf, self.nbf )
@@ -505,35 +526,3 @@ class Inverter(Direct, WuYang, ZMP, MRKS, OC, PDECO, Grider):
                 self.va += vxc_a.np
                 self.vb += vxc_b.np
         self.guide_potential_components = guide_potential_components
-
-    def finalize_energy(self):
-        """
-        Calculates energy contributions
-        """
-
-        target_one        = self.wfn.to_file()['floatvar']['ONE-ELECTRON ENERGY']
-        target_two        = self.wfn.to_file()['floatvar']['TWO-ELECTRON ENERGY']
-
-        energy_kinetic    = contract('ij,ij', self.T, (self.Da + self.Db))
-        energy_external   = contract('ij,ij', self.V, (self.Da + self.Db))
-        energy_hartree_a  = 0.5 * contract('ij,ji', self.J0[0] + self.J0[1], self.Da)
-        energy_hartree_b  = 0.5 * contract('ij,ji', self.J0[0] + self.J0[1], self.Db)
-
-        print("WARNING: XC Energy is not yet properly calculated")
-        energies = {"One-Electron Energy" : energy_kinetic + energy_external,
-                    "Two-Electron Energy" : energy_hartree_a + energy_hartree_b,
-                    # "XC"                  : energy_ks,
-                    # "Total Energy"        : energy_kinetic   + energy_external  + \
-                    #                         energy_hartree_a + energy_hartree_b + \
-                    #                         energy_ks 
-                    }
-
-        target_energies =  {"One-Electron Energy" : target_one,
-                            "Two-Electron Energy" : target_two,
-                           }
-
-        self.energies = energies
-        self.target_energies = target_energies
-
-        # print(f"Final Energies: {self.energies}")
-        # print(f"Target Energies: {self.target_energies}")
